/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 *
 * This video driver is STRONGLY derived from the video drivers of the
 * XFree 4.0.0 project
 */

#include "compil.h"
#include "neomagic.h"
#include "card.h"

struct NeoMagicID {
	unsigned value;
	const char* name;
	int supported;
	int mul_min;
	int mul_max;
	int div_min;
	int div_max;
	int p_min;
	int p_max;
	long ref; /* kHz */
	long vco_min; /* kHz */
	long vco_max; /* kHz */
	int clock_type;
};

/* Clock */
#define NEOMAGIC_CLOCK_128 1
#define NEOMAGIC_CLOCK_256 2

/* Configuration */
#define NEOMAGIC_CONF_UNSUPPORTED 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define NEOMAGIC_CONF_NM207X 1, 1, 128, 1, 32, 0, 1, 14318, 11000, 65000, NEOMAGIC_CLOCK_128
#define NEOMAGIC_CONF_NM209X 1, 1, 128, 1, 32, 0, 1, 14318, 11000, 80000, NEOMAGIC_CLOCK_128
#define NEOMAGIC_CONF_NM21XX 1, 1, 128, 1, 32, 0, 1, 14318, 11000, 90000, NEOMAGIC_CLOCK_128
#define NEOMAGIC_CONF_256 1, 1, 128, 1, 32, 0, 1, 14318, 11000, 110000, NEOMAGIC_CLOCK_256

static struct NeoMagicID NeoMagicIDList[] = {
{ 0x0001, "NeoMagic MagicGraph 128 (NM2070)", NEOMAGIC_CONF_NM207X },
{ 0x0002, "NeoMagic MagicGraph 128V (NM2090)", NEOMAGIC_CONF_NM209X },
{ 0x0003, "NeoMagic MagicGraph 128ZV (NM2093)", NEOMAGIC_CONF_NM209X },
{ 0x0083, "NeoMagic MagicGraph 128ZV+ (NM2097)", NEOMAGIC_CONF_NM209X },
{ 0x0004, "NeoMagic MagicGraph 128XD (NM2160)", NEOMAGIC_CONF_NM21XX },
{ 0x0005, "NeoMagic MagicMedia 256AV (NM2200)", NEOMAGIC_CONF_256 },
{ 0x0006, "NeoMagic MagicMedia 256ZX (NM2360)", NEOMAGIC_CONF_256 },
{ 0x0016, "NeoMagic MagicMedia 256XL+ (NM2380)", NEOMAGIC_CONF_256 },
{ 0, 0, NEOMAGIC_CONF_UNSUPPORTED }
};

static unsigned NeoMagicBusDeviceFunc;

static struct NeoMagicID* neomagic_card; /* NeoMagic detected */

const char* driverNeoMagic(void) {
	return neomagic_card->name;
}

int detectNeoMagic(void)
{
	int i;

	if (pci_detect()!=0) {
		CARD_LOG(("neomagic: PCI BIOS not Installed.\n"));
		return 0;
	}

	for(i=0;NeoMagicIDList[i].name;++i) {
		if (pci_find_device(0x000010c8,NeoMagicIDList[i].value,&NeoMagicBusDeviceFunc)==0)
			break;
	}

	if(!NeoMagicIDList[i].name) {
		CARD_LOG(("neomagic: not found\n"));
		return 0;
	}

	CARD_LOG(("neomagic: found %s, device id %04x\n", NeoMagicIDList[i].name, NeoMagicIDList[i].value));

	if (!NeoMagicIDList[i].supported) {
		CARD_LOG(("neomagic: card not supported\n"));
		return 0;
	}

	neomagic_card = NeoMagicIDList + i;

	return 1;
}

void resetNeoMagic(void)
{
}

static long NeoMagicGetVClock(long dotclock, int STACK_PTR* s_mul, int STACK_PTR* s_div, int STACK_PTR* s_p) {
	int best_mul,best_div,best_p;
	long best_clock,best_vco;
	int pll_mul,pll_div,pll_p;

	CARD_LOG(("neomagic: clock range is %.2f - %.2f MHz\n", (double)neomagic_card->vco_min / 1000 / (1 << neomagic_card->p_max), (double)neomagic_card->vco_max / 1000 / (1 << neomagic_card->p_min) ));

	best_clock = -1;
	best_mul = 0;
	best_div = 0;
	best_p = 0;
	best_vco = 0;

	for(pll_div=neomagic_card->div_min;pll_div<=neomagic_card->div_max;++pll_div)
	for(pll_p=neomagic_card->p_min;pll_p<=neomagic_card->p_max;++pll_p)
	for(pll_mul=neomagic_card->mul_min;pll_mul<=neomagic_card->mul_max;++pll_mul) {
		long vco = neomagic_card->ref * 1000L * pll_mul / pll_div;
		long new_clock = vco / (1 << pll_p);

		if (vco < neomagic_card->vco_min*1000L && pll_p!=neomagic_card->p_max)
			continue;
		if (vco > neomagic_card->vco_max*1000L)
			continue;

		if (best_clock < 0 || labs(new_clock - dotclock) < labs(best_clock - dotclock)) {
			best_mul = pll_mul;
			best_div = pll_div;
			best_p = pll_p;
			best_clock = new_clock;
			best_vco = vco;
		}
	}

	if (best_clock < 0)
		return -1;

	if (best_vco < neomagic_card->vco_min*1000)
		CARD_LOG(("neomagic: VCO clock %.2f is OUT of range %.2f - %.2f MHz\n", (double)best_vco / 1000000, (double)neomagic_card->vco_min / 1000, (double)neomagic_card->vco_max / 1000));
	else
		CARD_LOG(("neomagic: VCO clock %.2f is in range %.2f - %.2f MHz\n", (double)best_vco / 1000000, (double)neomagic_card->vco_min / 1000, (double)neomagic_card->vco_max / 1000));

	*s_mul = best_mul;
	*s_div = best_div;
	*s_p = best_p;

	CARD_LOG(("neomagic: ref:%.2f, mul:%d, div:%d, shift:%d\n",(double)neomagic_card->ref / 1000,best_mul,best_div,best_p));

	return best_clock;
}

static void SetNeoMagicPLL(void)
{
	BYTE d0;
	d0 = _in(0x3CC);
	d0 = d0 | 0x0C;
	_out(0x3C2, d0);
}

static void NeoMagicSetVClock(int pll_mul, int pll_div, int pll_p) {
	if (neomagic_card->clock_type == NEOMAGIC_CLOCK_128) {
		BYTE num = (pll_mul - 1) | (pll_p << 7);
		BYTE den = pll_div - 1;
		SetGraphReg(0x9B,num);
		SetGraphReg(0x9F, den);
	} else if (neomagic_card->clock_type == NEOMAGIC_CLOCK_256) {
		BYTE num_l = pll_mul - 1;
		BYTE num_h = pll_p << 7;
		BYTE den = pll_div - 1;
		BYTE d0;
		SetGraphReg(0x9B, num_l);
		d0 = GetGraphReg(0x8F);
		d0 &= 0x0F;
		d0 |= num_h & ~0x0F;
		SetGraphReg(0x8F, d0);
		SetGraphReg(0x9F, den);
	}

	SetSeqDivider(1);
	SetNeoMagicPLL();
}

static void NeoMagicUnlock(void) {
	VGAUnlock();
	SetGraphReg(0x09,0x26);
}

static void NeoMagicLock(void) {
	SetGraphReg(0x09,0x00);
	VGALock();
}

int setNeoMagic(void)
{
	STACK_DECL CRTCPARAM cp;
	int pll_mul,pll_div,pll_p;

	if(GetMode15KHz(&cp) == 0)
		return 0;

	if (cp.interlace) {
		CARD_LOG(("neomagic: interlaced modes not supported\n"));
		return 0;
	}

	CARD_LOG(("neomagic: clock requested:%.2f MHz\n", (double)cp.dotclockHz / 1000000 ));
	cp.dotclockHz = NeoMagicGetVClock(cp.dotclockHz,&pll_mul,&pll_div,&pll_p);
	if (cp.dotclockHz<0) {
		CARD_LOG(("neomagic: clock out of range\n"));
		return 0;
	}
	CARD_LOG(("neomagic: clock selected:%.2f MHz\n", (double)cp.dotclockHz / 1000000 ));

	NeoMagicUnlock();
	DisableGenerateSignal();
	SetCrtcVGARegister(&cp);
	NeoMagicSetVClock(pll_mul, pll_div, pll_p);
	SetDoubleScan(cp.doublescan);
	SetHVPolarity(cp.hpolarity, cp.vpolarity);
	EnableGenerateSignal();
	NeoMagicLock();

	return 1;
}
