/*
 * This file is part of the AdvanceVBE project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "types.h"
#include "lib.h"
#include "tsr.h"

/***************************************************************************/
/* TSR */

struct tsr_type {
	uint16 segment;     /* TSR segment */
	uint16 active;      /* !=0 TSR active */
	uint16 id;          /* TSR ID */

	int far (far * done)(void); /* deinstallation call */
	int far (far * init)(void); /* installation call */
};

struct tsr_type tsr_data;

/***************************************************************************/
/* RUThere */

/* Old interrupt */
extern void far * int_2f_old;

/* ID used for the interrupt trap */
extern uint8 int_2f_id;

// nuovo int
#ifdef __cplusplus
extern "C"
#endif
void int_2f(void);

#ifdef __cplusplus
extern "C"
#endif
void int_2f_trap(regs_16 far * r) {
	/* assert( r->b.ah == tsr_data.id) */
	r->x.cx = RUT_ACK1;
	r->x.dx = RUT_ACK2;
	r->x.es = _DS;
	r->x.bx = FP_OFF( &tsr_data );
	r->h.al = 0xFF;
}

void int_2f_call(void) {
	asm push si
	asm push di
	asm push bp
	asm push ds
	asm int 0x2f
	asm pop ds
	asm pop bp
	asm pop di
	asm pop si
}

int rut_free_code(unsigned code) {
	asm mov ah,[byte ptr code]
	asm xor al,al
	asm xor bx,bx
	asm xor cx,cx
	asm xor dx,dx
	int_2f_call();
	asm cmp al,0
	asm jne error
	return 1;
error:
	return 0;
}

// installa il rut
int rut_search(void) {
	unsigned i;
	tsr_data.id = 0;
	for(i=0xC0;i<=0xFF;++i) {
		if (rut_free_code(i)) {
			tsr_data.id = i;
			int_2f_id = i;
			return 1;
		}
	}
	return 0;
}

// test di installazione
// reg:
//   ah    code
//   al    0xFF
//   cx    RUT_ACK1
//   dx    RUT_ACK2
//   es:bx &tsr_type
struct tsr_type far * rut_is_installed_code(uint8 code) {
	asm mov ah,[code]
	asm xor al,al
	asm xor bx,bx
	asm xor cx,cx
	asm xor dx,dx
	int_2f_call();
	asm cmp al,0FFh
	asm jne notinst
	asm cmp cx,RUT_ACK1
	asm jne notinst
	asm cmp dx,RUT_ACK2
	asm jne notinst

	asm mov dx,es
	asm mov ax,bx
	return (struct tsr_type far*)MK_FP(_DX,_AX);
notinst:
	return 0;
}

// cerca il programma
// effect:
//   se trovato tsr punta alla zona dati
// return:
//   0 non trovato
//   !=0 puntatore alla zona dati del TSR
struct tsr_type far* tsr_is_installed(void) {
	unsigned i;
	for(i=0xC0;i<=0xFF;++i) {
		struct tsr_type far* p = rut_is_installed_code(i);
		if (p)
			return p;
	}
	return 0;
}

/***************************************************************************/
/* INT */

extern void far * int_10_old;
#ifdef __cplusplus
extern "C"
#endif
void int_10(void);

/***************************************************************************/
/* TSR */

// disinstallare il TSR
// return:
//   0 no
//   1 si
int far done(void) {
	if ( (int_10 == getvect(0x10))
		&& (int_2f == getvect(0x2f))) {
		//  possibile disinstallarlo

		tsr_data.active = 0;

		tsr_done_0();

		// int tsr
		setvect(0x10,int_10_old);
		setvect(0x2f,int_2f_old);

		tsr_done_1();

		return 1;
	}

	return 0;
}

// installa il TSR
// return:
//   0 non  stato possibile installare
//   1 ok
int far init(void) {

	(void far*)tsr_data.done = MK_FP( _CS, FP_OFF( done ));
	(void far*)tsr_data.init = MK_FP( _CS, FP_OFF( init ));

	tsr_data.active = 0;
	tsr_data.segment = _DS;

	if (!rut_search()) {
#if !defined(NDEBUG)
		cputs("msg: nessun int2f ID trovato libero\n\r");
#endif
	return 0;
	}

#if !defined(NDEBUG)
	cputs("ID "); cputu(tsr_data.id,2,'0',16); cputs("\n\r");
#endif

	if (!tsr_init_0()) {
		return 0;
	}

	// int tsr
	int_10_old = getvect(0x10);
	int_2f_old = getvect(0x2f);

	setvect(0x10,int_10);
	setvect(0x2f,int_2f);

	if (!tsr_init_1()) {
	setvect(0x10,int_10_old);
	setvect(0x2f,int_2f_old);

	tsr_done_0();
	return 0;
	}

	tsr_data.active = 1;

	return 1;
}

// chiama init del tsr in memoria
int mem_init(struct tsr_type far* tsr) {
	int far (far* func)(void);
#if !defined(NDEBUG)
	if (!tsr) {
		cputs("error: tsr is null!\n\r");
		return 0;
	}
#endif
	func = tsr->init;
	_AX = tsr->segment;
	asm push ds
	asm mov ds,ax
	func();
	asm pop ds
	return _AX;
}

// chiama done del tsr in memoria
int mem_done(struct tsr_type far* tsr) {
	int far (far*func)(void);
#if !defined(NDEBUG)
	if (!tsr) {
		cputs("error: tsr is null!\n\r");
		return 0;
	}
#endif
	func = tsr->done;
	_AX = tsr->segment;
	asm push ds
	asm mov ds,ax
	func();
	asm pop ds
	return _AX;
}

/***************************************************************************/
/* main */

#define EXIT_FAILURE 2
#define EXIT_SUCCESS 1
#define EXIT_TSR 0

int main(void) {
	struct tsr_type far * tsr = tsr_is_installed();

	if (tsr) {
		if (tsr->active) {
			if (mem_done(tsr)) {
#if !defined(NDEBUG)
				cputs("msg: tsr removed\n\r");
#endif
			} else {
				cputs("TSR can't be removed\n\r");
			}
		}
		if (tsr && tsr->active==0) {
			int e = freemem( tsr->segment );
			if (e != 0) {
				cputs("error: free mem\n\r");
				return EXIT_FAILURE;
			}
		}
		return EXIT_SUCCESS;
	} else {
#if !defined(NDEBUG)
		cputs("CS "); cputu(_CS,4,'0',16); cputs("\n\r");
#endif
		if (!init())
			return EXIT_FAILURE;

		freeenv();

		return EXIT_TSR;
	}
}