/*
 * This file is part of the AdvanceVBE project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "tsr.h"
#include "lib.h"

#include "card.h"
#include "s3.h"
#include "cirrus.h"
#include "s6326.h"
#include "matrox.h"
#include "savage.h"
#include "laguna.h"
#include "tdfx.h"
#include "vbe3.h"
#include "neomagic.h"

/***************************************************************************/
/* VBE */

/* OEM information */
#define OEM_VENDOR_STR "Andrea Mazzoleni"
#define OEM_PRODUCT_STR "AdvanceVBE"
#define OEM_VERSION_NUM 0x0001
#define OEM_VERSION_STR "0.1 " __DATE__
#define OEM_STR OEM_VENDOR_STR " " OEM_PRODUCT_STR " " OEM_VERSION_STR

#define VBE_CUSTOM_MODE 0x120

typedef struct vbe_VbeInfoBlock_struct {
	uint8 VESASignature[4] ; /* 'VESA' 4 byte signature */
	uint16 VESAVersion; /* VBE version number */
	char far* OemStringPtr; /* Pointer to OEM string */
	uint32 Capabilities; /* Capabilities of video card */
	uint16 far* VideoModePtr; /* Pointer to supported modes */
	uint16 TotalMemory;  /* Number of 64kb memory blocks */

	/* VBE 2.0 extension information */
	uint16 OemSoftwareRev; /* OEM Software revision number */
	char far* OemVendorNamePtr; /* Pointer to Vendor Name string */
	char far* OemProductNamePtr; /* Pointer to Product Name string */
	char far* OemProductRevPtr; /* Pointer to Product Revision str */
	uint8 Reserved[222]; /* Pad to 256 byte block size */
	uint8 OemDATA[256]; /* Scratch pad for OEM data */
} vbe_VbeInfoBlock;

/* Mode information block */
typedef struct vbe_ModeInfoBlock_struct {
	uint16 ModeAttributes; /* Mode attributes */
	uint8 WinAAttributes; /* Window A attributes */
	uint8 WinBAttributes; /* Window B attributes */
	uint16 WinGranularity; /* Window granularity in k */
	uint16 WinSize; /* Window size in k */
	uint16 WinASegment; /* Window A segment */
	uint16 WinBSegment; /* Window B segment */
	uint32 WinFuncPtr; /* Pointer to window function */
	uint16 BytesPerScanLine; /* Bytes per scanline */
	uint16 XResolution; /* Horizontal resolution */
	uint16 YResolution; /* Vertical resolution */
	uint8 XCharSize; /* Character cell width */
	uint8 YCharSize; /* Character cell height */
	uint8 NumberOfPlanes; /* Number of memory planes */
	uint8 BitsPerPixel; /* Bits per pixel */
	uint8 NumberOfBanks; /* Number of CGA style banks */
	uint8 MemoryModel; /* Memory model type */
	uint8 BankSize; /* Size of CGA style banks */
	uint8 NumberOfImagePages; /* Number of images pages */
	uint8 ReservedPage; /* Reserved */
	uint8 RedMaskSize; /* Size of direct color red mask */
	uint8 RedFieldPosition; /* Bit posn of lsb of red mask */
	uint8 GreenMaskSize; /* Size of direct color green mask */
	uint8 GreenFieldPosition; /* Bit posn of lsb of green mask */
	uint8 BlueMaskSize; /* Size of direct color blue mask */
	uint8 BlueFieldPosition; /* Bit posn of lsb of blue mask */
	uint8 RsvdMaskSize; /* Size of direct color res mask */
	uint8 RsvdFieldPosition; /* Bit posn of lsb of res mask */
	uint8 DirectColorModeInfo; /* Direct color mode attributes */

	/* VBE 2.0 extensions information */
	uint32 PhysBasePtr; /* Physical address for linear buf */
	uint32 OffScreenMemOffset; /* Pointer to start of offscreen mem */
	uint16 OffScreenMemSize; /* Amount of offscreen mem in 1K's */

	/* VBE 3.0 extensions */
	uint16 LinBytesPerScanLine; /* Bytes per scanline */
	uint8 BnkNumberOfPages; /* Number of images pages (banked) */
	uint8 LinNumberOfPages; /* Number of images pages (linear) */
	uint8 LinRedMaskSize; /* Size of direct color red mask */
	uint8 LinRedFieldPosition; /* Bit posn of lsb of red mask */
	uint8 LinGreenMaskSize; /* Size of direct color green mask */
	uint8 LinGreenFieldPosition; /* Bit posn of lsb of green mask */
	uint8 LinBlueMaskSize; /* Size of direct color blue mask */
	uint8 LinBlueFieldPosition;    /* Bit posn of lsb of blue mask */
	uint8 LinRsvdMaskSize; /* Size of direct color res mask */
	uint8 LinRsvdFieldPosition;    /* Bit posn of lsb of res mask */
	uint32 MaxPixelClock; /* Maximum pixel clock */

	uint8 Reserved[190]; /* Pad to 256 byte block size */
} vbe_ModeInfoBlock;

/* Flags for combining with video modes during mode set */
#define vbeDontClear 0x8000 /* Dont clear display memory */
#define vbeLinearBuffer 0x4000 /* Enable linear framebuffer mode */
#define vbeRefreshCtrl 0x0800 /* Use refresh rate control */
#define vbeModeMask 0x03FF /* Mask for VBE mode numbers */

/* Flags for the mode attributes returned by VBE_getModeInfo. If */
/* vbeMdNonBanked is set to 1 and vbeMdLinear is also set to 1, then only */
/* the linear framebuffer mode is available. */
#define vbeMdAvailable 0x0001 /* Video mode is available */
#define vbeMdTTYOutput 0x0004 /* TTY BIOS output is supported */
#define vbeMdColorMode 0x0008 /* Mode is a color video mode */
#define vbeMdGraphMode 0x0010 /* Mode is a graphics mode */
#define vbeMdNonVGA 0x0020 /* Mode is not VGA compatible */
#define vbeMdNonBanked 0x0040 /* Banked mode is not supported */
#define vbeMdLinear 0x0080 /* Linear mode supported */
#define vbeMdDoubleScan 0x0100 /* Double scan mode supported */
#define vbeMdInterlace 0x0200 /* Iterlaced mode supported */
#define vbeMdTripleBuffer 0x0400 /* Hardware triple buffering support */
#define vbeMdStereoscopic 0x0800 /* Hardware stereoscopic display support */

/* VBE 3.0 CRTC Info Block */
typedef struct vbe_CRTCInfoBlock_struct {
	uint16 HorizontalTotal; /* Horizontal total in pixels */
	uint16 HorizontalSyncStart; /* Horizontal sync start in pixels */
	uint16 HorizontalSyncEnd; /* Horizontal sync end in pixels */
	uint16 VerticalTotal; /* Vertical total in lines */
	uint16 VerticalSyncStart; /* Vertical sync start in lines */
	uint16 VerticalSyncEnd; /* Vertical sync end in lines */
	uint8 Flags; /* Flags (Interlaced, Double Scan etc) */
	uint32 PixelClock; /* Pixel clock in units of Hz */
	uint16 RefreshRate; /* Refresh rate in units of 0.01 Hz */
	uint8 Reserved[40]; /* Remainder of ModeInfoBlock */
} vbe_CRTCInfoBlock;

/* Flags for CRTC Flags */
#define vbeCRTCDoubleScan 0x01 /* Enable double scanned mode */
#define vbeCRTCInterlaced 0x02 /* Enable interlaced mode */
#define vbeCRTCHorizontalSyncNegative 0x04 /* Horizontal sync is negative */
#define vbeCRTCVerticalSyncNegative 0x08 /* Vertical sync is negative */

/* VBE version */
static unsigned vbe_version;

int vbe_detect(vbe_VbeInfoBlock STACK_PTR* info) {
	regs_16 regs;

	memcpy(&info->VESASignature,"VBE2",4);

	regs_16_zero(&regs);
	regs.x.ax = 0x4f00;
	regs.x.es = FP_SEG(info);
	regs.x.di = FP_OFF(info);
	int_10_call(&regs);

	if (regs.x.ax != 0x004f)
		return -1;

	vbe_version = info->VESAVersion;

	return 0;
}

int vbe_mode_info_get(vbe_ModeInfoBlock STACK_PTR* info, unsigned mode) {
	regs_16 regs;

	regs_16_zero(&regs);
	regs.x.ax = 0x4f01;
	regs.x.cx = mode & vbeModeMask;
	regs.x.es = FP_SEG(info);
	regs.x.di = FP_OFF(info);
	int_10_call(&regs);
	if (regs.x.ax != 0x004f)
		return -1;

	return 0;
}

int vbe_scanline_set(unsigned bytes_per_scanline, unsigned pixel_per_scanline) {
	regs_16 regs;

	regs_16_zero(&regs);
	regs.x.ax = 0x4f06;
	if (vbe_version >= 0x200) {
		regs.h.bl = 2;
		regs.x.cx = bytes_per_scanline;
	} else {
		regs.h.bl = 0;
		regs.x.cx = pixel_per_scanline;
	}
	int_10_call(&regs);
	if (regs.x.ax != 0x004f)
		return -1;

	return 0;
}

/***************************************************************************/
/* ADVVBE */

#define MODE_MAX 16

typedef struct mode_info_struct {
	CRTCPARAM param;
	unsigned effective_mode;
	unsigned reported_mode;
	unsigned bytes_per_scanline;
	int require_set_scanline_size;
} mode_info;

static mode_info mode_map[MODE_MAX];

/* number of video mode defined */
static unsigned mode_max;

/* active video mode */
static mode_info* mode_active;

typedef struct driver_info_struct {
	const char* name;
	int require_vbe3;
	int (*detect)(void);
	const char* (*driver)(void);
	int (*set)(void);
	void (*reset)(void);
} driver_info;

static driver_info driver_map[]=
{
//	{"laguna", 0, detectLaguna, driverLaguna, setLaguna, resetLaguna}, /* Laguna driver */
	{"3dfx", 0, detect3dfx, driver3dfx, set3dfx, reset3dfx}, /* 3dfx driver */
	{"savage", 0, detectSavage, driverSavage, setSavage, resetSavage}, /* Savage driver */
	{"sis6326", 0, detectSiS6326, driverSiS6326, setSiS6326, resetSiS6326}, /* SiS6326 driver */
//	{"ati", 0, detectATI, driverATI, setATI, resetATI} , /* ATI driver */
//	{"matrox", 0, detectMatrox, driverMatrox, setMatrox, resetMatrox}, /* Matrox driver */
//	{"r128", 0, detectR128, driverR128, setR128, resetR128}, /* R128 driver */
	{"neomagic", 0, detectNeoMagic, driverNeoMagic, setNeoMagic, resetNeoMagic}, /* NeoMagic driver */
	{"s3", 0, detectS3, driverS3, setS3, resetS3}, /* S3 driver */
	{"cirrus", 0, detectCirrus, driverCirrus, setCirrus, resetCirrus}, /* Cirrus driver */
	{"vbe3", 1, detectVBE3, driverVBE3, setVBE3, resetVBE3}, /* VBE3 driver */
	{0,0,0,0}
};

/* active video driver */
static driver_info* driver_active;

/* vbe video mode list */
static uint16 vbe_mode_list[MODE_MAX+1];

static char driver_name[16];

static char far* gettok_ptr = 0;

int _pascal gettok_separator(char c) {
	return c==' ' || c=='\t' || c=='=' || c=='(' || c==')';
}

char far* gettok(void) {
	char far* r;
	while (gettok_separator(*gettok_ptr))
		++gettok_ptr;
	if (!*gettok_ptr)
		return 0;
	r = gettok_ptr;
	while (*gettok_ptr && !gettok_separator(*gettok_ptr))
		++gettok_ptr;
	if (*gettok_ptr) {
		*gettok_ptr = 0;
		++gettok_ptr;
	}
	return r;
}

struct mode_number_struct {
	unsigned number;
	unsigned x;
	unsigned y;
	unsigned bits;
};

/* VESA 1.2 default numbering */
static struct mode_number_struct mode_number_map[] = {
{ 0x100,640,400,8 },
{ 0x101,640,480,8 },
{ 0x103,800,600,8 },
{ 0x105,1024,768,8 },
{ 0x107,1280,1024,8 },
{ 0x10D,320,200,15 },
{ 0x10E,320,200,16 },
{ 0x10F,320,200,24 },
{ 0x110,640,480,15 },
{ 0x111,640,480,16 },
{ 0x112,640,480,24 },
{ 0x113,800,600,15 },
{ 0x114,800,600,16 },
{ 0x115,800,600,24 },
{ 0x116,1024,768,15 },
{ 0x117,1024,768,15 },
{ 0x118,1024,768,24 },
{ 0x119,1280,1024,15 },
{ 0x11A,1280,1024,16 },
{ 0x11B,1280,1024,24 },
{0,0,0,0}
};

static unsigned mode_number_generator = VBE_CUSTOM_MODE;

unsigned mode_find_number(unsigned x, unsigned y, unsigned bits) {
	unsigned i;
	unsigned number;
	for(i=0;mode_number_map[i].number;++i)
		if (x==mode_number_map[i].x && mode_number_map[i].y==y && mode_number_map[i].bits==bits)
			break;
	if (!mode_number_map[i].number)
		return mode_number_generator++;
	number = mode_number_map[i].number;
	for(i=0;i<mode_max;++i)
		if (mode_map[i].reported_mode == number)
			break;
	if (i<mode_max)
		return mode_number_generator++;
	return number;
}

int mode_load(void) {
	STACK_DECL char buffer[128];

	int handle = open("vbe.cfg",O_RDONLY);
	if (handle == -1) {
		cputs("Configuration file vbe.cfg not found\r\n");
		return 0;
	}

	mode_max = 0;

	while (gets(handle, buffer, sizeof(buffer))) {
		STACK_DECL vbe_ModeInfoBlock info;
		char far* s;
		mode_info far* mode_ptr = mode_map + mode_max;

		/* too many modes */
		if (mode_max == MODE_MAX) {
			close(handle);
			cputs("Too many modelines\r\n");
			return 0;
		}

		gettok_ptr = buffer;

		memset(mode_ptr,0,sizeof(mode_info));

		s = gettok();
		if (!s) continue;

		if (strcmp(s,"vbeline_driver")==0) {
			s = gettok();
			if (!s) continue;

			strncpy(driver_name,s,sizeof(driver_name));
			driver_name[sizeof(driver_name)-1] = 0;
			continue;
		}

		if (memcmp(s,"modeline",8)!=0)
			continue;

		/* skip the name */
		s = gettok();
		if (!s) continue;

		s = gettok();
		if (!s) continue;
		if (strcmp(s,"vbeline")!=0)
			continue;

		s = gettok();
		if (!s) continue;
		mode_ptr->effective_mode = strtol(s,16) & vbeModeMask;

		s = gettok();
		if (!s) continue;
		mode_ptr->param.dotclockHz = strtold(s,1000000L);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HDisp = mode_ptr->param.width = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HSStart = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HSEnd = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.HTotal = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VDisp = mode_ptr->param.height = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VSStart = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VSEnd = strtol(s,10);

		s = gettok();
		if (!s) continue;
		mode_ptr->param.VTotal = strtol(s,10);

		while ((s = gettok())!=0) {
			if (strcmp(s,"doublescan")==0)
				mode_ptr->param.doublescan = 1;
			if (strcmp(s,"interlaced")==0)
				mode_ptr->param.interlace = 1;
			if (strcmp(s,"-hsync")==0)
				mode_ptr->param.hpolarity = 1;
			if (strcmp(s,"-vsync")==0)
				mode_ptr->param.vpolarity = 1;
		}

		/* convert the format to the requested */
		if (mode_ptr->param.doublescan) {
			mode_ptr->param.VDisp *= 2;
			mode_ptr->param.VBStart *= 2;
			mode_ptr->param.VSStart *= 2;
			mode_ptr->param.VSEnd *= 2;
			mode_ptr->param.VBEnd *= 2;
			mode_ptr->param.VTotal *= 2;
		}
		if (mode_ptr->param.interlace) {
			mode_ptr->param.VDisp /= 2;
			mode_ptr->param.VBStart /= 2;
			mode_ptr->param.VSStart /= 2;
			mode_ptr->param.VSEnd /= 2;
			mode_ptr->param.VBEnd /= 2;
			mode_ptr->param.VTotal /= 2;
		}

		if (vbe_mode_info_get(&info,mode_ptr->effective_mode)!=0) {
			close(handle);
			cputs("VBE mode ");
			cputu(mode_ptr->effective_mode,0,' ',16);
			cputs(" not supported by your current VBE BIOS\r\n");
			return 0;
		}

		mode_ptr->param.interlaceratio = 50;
		mode_ptr->param.bits = info.BitsPerPixel;
		mode_ptr->param.original_width = info.XResolution;
		mode_ptr->param.original_height = info.YResolution;

		/* blanck */
		mode_ptr->param.HBStart = mode_ptr->param.HDisp + 1;
		if (mode_ptr->param.HBStart > mode_ptr->param.HSStart)
			mode_ptr->param.HBStart = mode_ptr->param.HSStart;
		mode_ptr->param.HBEnd = mode_ptr->param.HSEnd + 1;
		mode_ptr->param.VBStart = mode_ptr->param.VDisp + 1;
		if (mode_ptr->param.VBStart > mode_ptr->param.VSStart)
			mode_ptr->param.VBStart = mode_ptr->param.VSStart;
		mode_ptr->param.VBEnd = mode_ptr->param.VSEnd + 1;

		/* set the scanline size */
		if (mode_ptr->param.width > info.XResolution) {
			unsigned bytes_per_pixel = (info.BitsPerPixel + 7) / 8;
			mode_ptr->bytes_per_scanline = (info.XResolution * bytes_per_pixel + 0xF) & ~0xF;
			mode_ptr->require_set_scanline_size = 1;
		} else {
			mode_ptr->bytes_per_scanline = info.BytesPerScanLine;
			mode_ptr->require_set_scanline_size = 0;
		}

		vbe_mode_list[mode_max] =
		mode_ptr->reported_mode = mode_find_number(mode_ptr->param.width,mode_ptr->param.height,mode_ptr->param.bits);

		++mode_max;
	}

	close(handle);

	vbe_mode_list[mode_max] = 0xFFFF;

	if (!mode_max) {
		cputs("No modelines found\r\n");
		return 0;
	}

	return 1;
}

void mode_print(void) {
	unsigned i;

	cputs("\r\nAvaliable modes:\r\n");
	for(i=0;i<mode_max;++i) {
		unsigned v;

		cputs("        ");
		cputu(mode_map[i].reported_mode,4,' ',16);
		cputs("h ");
		cputu(mode_map[i].param.width,4,' ',10);
		cputs("x");
		cputu(mode_map[i].param.height,4,' ',10);
		cputs(" ");
		cputu(mode_map[i].param.bits,2,' ',10);
		cputs(" bit ");

		v = mode_map[i].param.dotclockHz / (mode_map[i].param.HTotal * 100L);
		cputu(v / 10,2,' ',10);
		cputs(".");
		cputu(v % 10,1,' ',10);
		cputs(" kHz/");

		v = 10L * mode_map[i].param.dotclockHz / (mode_map[i].param.HTotal * (long)mode_map[i].param.VTotal);
		cputu(v / 10,2,' ',10);
		cputs(".");
		cputu(v % 10,1,' ',10);
		cputs(" Hz ");

		cputs("\r\n");
	}
}

int vbe_init(void) {
	STACK_DECL vbe_VbeInfoBlock info;

#ifndef NDEBUG
	cputs("SP: ");
	cputu(_SP,4,'0',16);
	cputs("\r\n");
	cputs("DS: ");
	cputu(_DS,4,'0',16);
	cputs("\r\n");
#endif

	if (vbe_detect(&info)!=0) {
		cputs("VBE BIOS not found!\r\n");
		return 0;
	}

	if (!mode_load()) {
		return 0;
	}

	for(driver_active = driver_map;driver_active->name;++driver_active) {
		if (!driver_active->require_vbe3 || vbe_version >= 0x300) {
			if (!*driver_name || strcmp(driver_name,driver_active->name)==0 || strcmp(driver_name,"auto")==0) {
				if (driver_active->detect()) {
					break;
				}
			}
		}
	}

	if (!driver_active->name) {
		driver_active = 0;
		if (*driver_name) {
			cputs("Video board not supported\r\nTry removing the vbeline_driver option from the configuration file\r\n");
		} else {
			cputs("Video board not supported\r\n");
		}
		return 0;
	}

#ifndef NDEBUG
	cputs("SP: ");
	cputu(_SP,4,'0',16);
	cputs("\r\n");
	cputs("DS: ");
	cputu(_DS,4,'0',16);
	cputs("\r\n");
#endif

	return 1;
}

void vbe_done(void) {
	if (driver_active)
		driver_active->reset();
	driver_active = 0;
}

void far* oem_alloc(uint8 far * far * base, void far* data, unsigned size) {
	uint8 far* r = *base;
	memcpy(r,data,size);
	*base += size;
	return r;
}

void vbe_service_4f00(regs_16 far * regs) {
	vbe_VbeInfoBlock far * info = MK_FP(regs->x.es,regs->x.di);
	int vbe2 = memcmp(info->VESASignature,"VBE2",4)==0;
	int_10_call(regs);
	if (regs->x.ax == 0x004f) {
		uint8 far* base_data = info->OemDATA;
		uint8 far* base_reserved = info->Reserved;
		if (vbe2) {
			info->VideoModePtr = (uint16 far*)oem_alloc(&base_reserved,&vbe_mode_list,sizeof(vbe_mode_list));
			info->OemStringPtr = (char far*)oem_alloc(&base_data,OEM_STR,strlen(OEM_STR)+1);
			if (info->VESAVersion >= 0x200) {
				info->OemSoftwareRev = OEM_VERSION_NUM;
				info->OemVendorNamePtr = (char far*)oem_alloc(&base_data,OEM_VENDOR_STR,strlen(OEM_VENDOR_STR)+1);
				info->OemProductNamePtr = (char far*)oem_alloc(&base_data,OEM_PRODUCT_STR,strlen(OEM_PRODUCT_STR)+1);
				info->OemProductRevPtr = (char far*)oem_alloc(&base_data,OEM_VERSION_STR,strlen(OEM_VERSION_STR)+1);
			}
		} else {
			info->VideoModePtr = (uint16 far*)&vbe_mode_list;
			info->OemStringPtr = OEM_STR;
		}
		/* max is VBE 2.0 */
		if (info->VESAVersion > 0x200)
			info->VESAVersion = 0x200;
	}
}

void vbe_service_4f01(regs_16 far* regs) {
	vbe_ModeInfoBlock far* info;
	mode_info* mode_ptr;

	/* VGA mode */
	if ((regs->x.cx & 0x100) == 0) {
		int_10_call(regs);
		return;
	}

	for(mode_ptr = mode_map;mode_ptr!=mode_map + mode_max;++mode_ptr)
		if (mode_ptr->reported_mode == regs->x.cx)
			break;
	if (mode_ptr == mode_map + mode_max) {
		regs->x.ax = 0x014f;
		return;
	}

	info = MK_FP(regs->x.es,regs->x.di);

	regs->x.cx = mode_ptr->effective_mode;
	int_10_call(regs);
	regs->x.cx = mode_ptr->reported_mode;

	if (regs->x.ax != 0x004f)
		return;

	info->XResolution = mode_ptr->param.width;
	info->YResolution = mode_ptr->param.height;
	info->BytesPerScanLine = mode_ptr->bytes_per_scanline;
}

int vbe_service_4f02_preset(vbe_ModeInfoBlock STACK_PTR* info, vbe_CRTCInfoBlock STACK_PTR* crtc) {
	STACK_DECL CRTCPARAM cp;

	if (!GetMode15KHz(&cp))
		return -1;

	if ((info->ModeAttributes & vbeMdAvailable) == 0) {
		CARD_LOG(("vbeline: VBE report that the selected mode isn't avaliable\n"));
		return -1;
	}

	if (cp.interlace && (info->ModeAttributes & vbeMdInterlace)==0) {
		CARD_LOG(("vbeline: VBE report that the selected mode doesn't support interlace\n"));
		return -1;
	}

	if (cp.doublescan && (info->ModeAttributes & vbeMdDoubleScan)==0) {
		CARD_LOG(("vbeline: VBE report that the selected mode doesn't support doublescan\n"));
		return -1;
	}

	if (cp.interlace && cp.doublescan) {
		cputs("DDD");
		CARD_LOG(("vbeline: VBE doesn't support interlaced and doublescan at the same time\n"));
		return -1;
	}

	memset(crtc,0,sizeof(vbe_CRTCInfoBlock));

	/* if the resulting mode is smaller and if later sync values are set */
	/* use fake value to prevent a possible mode set failure if the */
	/* specified values are smaller than the fixed resolution */

	if (cp.HDisp < info->XResolution) {
		/* use fake values, correct values are set later */
		crtc->HorizontalSyncStart = info->XResolution * 41 / 40;
		crtc->HorizontalSyncEnd = info->XResolution * 10 / 9;
		crtc->HorizontalTotal = info->XResolution * 5 / 4;
	} else {
		crtc->HorizontalTotal = cp.HTotal;
		crtc->HorizontalSyncStart = cp.HSStart;
		crtc->HorizontalSyncEnd = cp.HSEnd;
	}
	if (cp.VDisp < info->YResolution) {
		/* use fake values, correct values are set later */
		crtc->VerticalSyncStart = info->YResolution * 31 / 30;
		crtc->VerticalSyncEnd = info->YResolution * 21 / 20;
		crtc->VerticalTotal = info->YResolution * 16 / 15;

		/* VBE3 requires doubled values for doublescan mode */
		/* but NOT half values for interlaced modes */
		if (cp.doublescan) {
			crtc->VerticalTotal *= 2;
			crtc->VerticalSyncStart *= 2;
			crtc->VerticalSyncEnd *= 2;
		}
	} else {
		crtc->VerticalTotal = cp.VTotal;
		crtc->VerticalSyncStart = cp.VSStart;
		crtc->VerticalSyncEnd = cp.VSEnd;

		/* VBE3 requires doubled values for doublescan mode */
		/* but NOT half values for interlaced modes */
		if (cp.interlace) {
			crtc->VerticalTotal *= 2;
			crtc->VerticalSyncStart *= 2;
			crtc->VerticalSyncEnd *= 2;
		}
	}

	crtc->Flags = 0;
	if (cp.doublescan)
		crtc->Flags |= vbeCRTCDoubleScan;
	if (cp.interlace)
		crtc->Flags |= vbeCRTCInterlaced;
	if (cp.hpolarity)
		crtc->Flags |= vbeCRTCHorizontalSyncNegative;
	if (cp.vpolarity)
		crtc->Flags |= vbeCRTCVerticalSyncNegative;

	crtc->PixelClock = cp.dotclockHz;
	crtc->RefreshRate = (long)crtc->PixelClock * 100 / ((long)cp.HTotal * cp.VTotal);

	CARD_LOG(("vbeline: mode_set_vbe3 %dx%d (%04x %.2f %d %d %d %d %d %d)\n",(unsigned)info->XResolution,(unsigned)info->YResolution,(unsigned)mode,(double)crtc->PixelClock/1E6,(unsigned)crtc->HorizontalSyncStart,(unsigned)crtc->HorizontalSyncEnd,(unsigned)crtc->HorizontalTotal,(unsigned)crtc->VerticalSyncStart,(unsigned)crtc->VerticalSyncEnd,(unsigned)crtc->VerticalTotal));

	return 0;
}

void vbe_service_4f02(regs_16 far* regs) {
	mode_info* mode_ptr;
	unsigned req_mode;
	unsigned req_flag;

	/* VGA mode */
	if ((regs->x.bx & 0x100) == 0) {
		mode_active = 0;
		int_10_call(regs);
		return;
	}

	req_mode = regs->x.bx & vbeModeMask;
	req_flag = regs->x.bx & ~vbeModeMask;

	for(mode_ptr = mode_map;mode_ptr!=mode_map + mode_max;++mode_ptr)
		if (mode_ptr->reported_mode == req_mode)
			break;
	if (mode_ptr == mode_map + mode_max) {
		regs->x.ax = 0x014f;
		return;
	}

	/* active mode */
	mode_active = mode_ptr;

	if (driver_active->require_vbe3) {
		STACK_DECL vbe_CRTCInfoBlock crtc;
		STACK_DECL vbe_ModeInfoBlock info;

		if (vbe_mode_info_get(&info,mode_active->effective_mode)!=0) {
			mode_active = 0;
			regs->x.ax = 0x014f;
			return;
		}

		if (vbe_service_4f02_preset(&info,&crtc)!=0) {
			mode_active = 0;
			regs->x.ax = 0x014f;
			return;
		}

		regs->x.bx = mode_active->effective_mode | req_flag | vbeRefreshCtrl;
		regs->x.es = FP_SEG(&crtc);
		regs->x.di = FP_OFF(&crtc);

		int_10_call(regs);
	} else {
		regs->x.bx = mode_active->effective_mode | req_flag;
		int_10_call(regs);
	}

	regs->x.bx = req_mode | req_flag;

	if (regs->x.ax != 0x004f) {
		mode_active = 0;
		return;
	}

	if (mode_active->require_set_scanline_size) {
		if (vbe_scanline_set(mode_active->bytes_per_scanline,mode_active->param.width)!=0) {
			mode_active = 0;
			regs->x.ax = 0x014f;
			return;
		}
	}

	driver_active->set();
}

void vbe_service_4f03(regs_16 far * regs) {
	int_10_call(regs);
	if (regs->x.ax != 0x004f)
		return;

	if (mode_active && regs->x.bx == mode_active->effective_mode)
		regs->x.bx = mode_active->reported_mode;
}

#define VBE_SERVICE_CALL_MAX 16

static unsigned long vbe_service_call[VBE_SERVICE_CALL_MAX];

const char* vbe_service_name[VBE_SERVICE_CALL_MAX] = {
/* 00 */ "Return VBE Controller Information",
/* 01 */ "Return VBE Mode Information",
/* 02 */ "Set VBE Mode",
/* 03 */ "Return current VBE Mode",
/* 04 */ "Save/Restore state",
/* 05 */ "Display Window Control",
/* 06 */ "Set/Get Logical Scan Line Length",
/* 07 */ "Set/Get Display Start",
/* 08 */ "Set/Get DAC Palette Format",
/* 09 */ "Set/Get Palette Data",
/* 0A */ "Return VBE Protected Mode Interface",
/* 0B */ /* "Get/Set Pixel Clock" */ 0, /* Only VBE3 */
0,
0,
0,
0
};

void int_10_trap(regs_16 far * regs) {
	if (regs->h.al < VBE_SERVICE_CALL_MAX)
		++vbe_service_call[regs->h.al];
	switch (regs->x.ax) {
		case 0x4f00 : vbe_service_4f00(regs); break;
		case 0x4f01 : vbe_service_4f01(regs); break;
		case 0x4f02 : vbe_service_4f02(regs); break;
		case 0x4f03 : vbe_service_4f03(regs); break;
		default:
			int_10_call(regs);
	}
}

/***************************************************************************/
/* Interface */

int GetMode15KHz(CRTCPARAM STACK_PTR *cp) {
	memcpy(cp,&mode_active->param,sizeof(CRTCPARAM));
	return 1;
}

/***************************************************************************/
/* TSR */

uint16 _pascal tsr_init_0(void) {
	return 1;
}

uint16 _pascal tsr_init_1(void) {
	cputs(OEM_PRODUCT_STR " by " OEM_VENDOR_STR " v" OEM_VERSION_STR "\r\n");

	if (!vbe_init())
		return 0;

	cputs("\r\nUsing driver `");
	cputs(driver_active->name);
	cputs("' for `");
	cputs(driver_active->driver());
	cputs("'\r\n");

	mode_print();

	if (vbe_version < 0x200) {
		cputs("\r\nYou have a VBE version prior of 2.0. Probably you want\r\n");
		cputs("to install a software VBE BIOS before AdvanceVBE\r\n");
	}

	return 1;
}

void _pascal tsr_done_0(void) {
}

void _pascal tsr_done_1(void) {
	unsigned i;

	vbe_done();

	cputs(OEM_PRODUCT_STR " uninstalled");

	for(i=0;i<VBE_SERVICE_CALL_MAX;++i)
		if (vbe_service_call[i])
			break;

	if (i < VBE_SERVICE_CALL_MAX) {
		cputs("\r\n\r\nVBE calls:\r\n");
		for(i=0;i<VBE_SERVICE_CALL_MAX;++i) {
			if (vbe_service_call[i] || vbe_service_name[i]) {
				cputs("        4F");
				cputu(i,2,'0',16);
				cputs("h ");
				cputu(vbe_service_call[i],6,' ',10);
				if (vbe_service_name[i]) {
					cputs(" ");
					cputs(vbe_service_name[i]);
				}
				cputs("\r\n");
			}
		}
	}
}